;;; Class: AB5080-V
;;; Title: LISP Your MESH to Shape!
;;; Speaker: Reinaldo Togores, University of Cantabria
;;; http://www.togores.net
;;; Document: LISP code file

;;; Listing 1. Function that sets a custom visual style.
(defun var-vis  ()
  (setvar "VSMONOCOLOR" " RGB:211,76,3")
  (setvar "VSFACECOLORMODE" 1)
  (setvar "VSEDGES" 1)
  (setvar "VSEDGECOLOR" "RGB:255,212,82")
  (setvar "VSISOONTOP" 0)
  (setvar "VSFACESTYLE" 2)
  (setvar "VSOBSCUREDEDGES" 0)
  (setvar "VSOCCLUDEDEDGES" 0)
  (setvar "PERSPECTIVE" 1))

;;; Listing 2 Sets the SW isometric view and Visual Style.
(defun cmd-SW ()
  (command "_dview" "" "_ca" 45 -135 "")
  (command "_zoom" "_e")
  (var-vis))

;;; Listing 3 Creating a mesh BOX using the MESH command.
(defun cmd-mesh-box
       (center dim-x dim-y dim-z div-x div-y div-z /)
  (setvar "DIVMESHBOXLENGTH" div-x)
  (setvar "DIVMESHBOXWIDTH" div-y)
  (setvar "DIVMESHBOXHEIGHT" div-z)
  (vl-cmdf "._MESH" "_B" "_C" center "_L" dim-x dim-y dim-z))

;;;Creating a planar square MESH using the EDGESURF command.
(defun cmd-square-mesh  (ins-pt side div / x-min x-max y-min y-max z edges)
  (setvar "MESHTYPE" 1)                                   ; MESH type object
  (setvar "SURFTAB1" div)                                 ; U tesselations
  (setvar "SURFTAB2" div)                                 ; V tesselations
  (setq x-min (nth 0 ins-pt)
        x-max (+ x-min side)
        y-min (nth 1 ins-pt) 
        y-max (+ y-min side)
        z     (nth 2 ins-pt))
  (vl-cmdf "_LINE" ins-pt (list x-max y-min z) "")               ; Line 1
  (setq edges (cons (entlast) edges)) 
  (vl-cmdf "_LINE" (list x-max y-min z) (list x-max y-max z) "") ; Line 2
  (setq edges (cons (entlast) edges))
  (vl-cmdf "_LINE" (list x-max y-max z) (list x-min y-max z) "") ; Line 3
  (setq edges (cons (entlast) edges))
  (vl-cmdf "_LINE" (list x-min y-max z) ins-pt "")               ; Line 4
  (setq edges (cons (entlast) edges))
  (apply 'vl-cmdf (cons "EDGESURF" edges))                ; Create MESH
  (mapcar 'entdel edges)                                  ; Erase lines
  (princ))

;;; ENTMAKing a MESH

(defun ent-mesh  (vertices faces edges / header ent-list)
  ;;Entity header:
  (setq header '((0 . "MESH")
                 (100 . "AcDbEntity")
                 (100 . "AcDbSubDMesh")
                 (71 . 2)
                 (72 . 0)
                 (91 . 0)))
  ;;Vertices data:
  (setq ent-list (cons (cons 92 (length vertices)) ent-list))
  (foreach vertex  vertices
    (setq ent-list (cons (cons 10 vertex) ent-list)))
  ;;Faces data:
  (setq
    ent-list (cons (cons 93
                         (+ (length faces)
                            (length (apply 'append faces))))
                   ent-list))
  (foreach face  faces
    (setq ent-list (cons (cons 90 (length face)) ent-list))
    (foreach vertexindex  face
      (setq ent-list (cons (cons 90 vertexindex) ent-list))))
  ;;Edges data:
  (setq ent-list (cons (cons 94 (/ (length edges) 2)) ent-list))
  (foreach vertexindex  edges
    (setq ent-list (cons (cons 90 vertexindex) ent-list)))
  ;;Crease data:
  (setq ent-list (cons (cons 95 (/ (length edges) 2)) ent-list))
  (setq ent-list (append header (reverse ent-list)))
  (entmake ent-list))


(defun C:TETRAHEDRON  (/ radius vertices faces edges)
  (setq radius   (getdist '(0 0 0) "\nCircumscribed sphere radius: ")
        vertices '((0 0 1)
                   (0 0.94280904 -0.33333333)
                   (-0.81649658 -0.47140452 -0.33333333)
                   (0.81649658 -0.47140452 -0.33333333))
        faces    '((0 1 2) (0 2 3) (0 3 1) (1 3 2))
        edges    '(0 1 1 2 2 0 2 3 3 0 3 1))
  (ent-mesh vertices faces edges)
  (command "_scale" (entlast) "" '(0 0 0) radius)
  (cmd-SW))


(defun C:HEXAHEDRON  (/ radius vertices faces edges)
  (setq radius   (getdist '(0 0 0) "\nCircumscribed sphere radius: ")
        vertices '((-0.57735 -0.57735 -0.57735)
                   (-0.57735 0.57735 -0.57735)
                   (0.57735 0.57735 -0.57735)
                   (0.57735 -0.57735 -0.57735)
                   (-0.57735 -0.57735 0.57735)
                   (-0.57735 0.57735 0.57735)
                   (0.57735 0.57735 0.57735)
                   (0.57735 -0.57735 0.57735))
        faces    '((0 1 2 3)
                   (4 5 1 0)
                   (5 6 2 1)
                   (6 7 3 2)
                   (7 4 0 3)
                   (7 6 5 4))
        edges    '(0 1 1 2 2 3 3 0 4 5 5 1 0 4 5 6 6 2 6 7 7 3 7 4))
  (ent-mesh vertices faces edges)
  (command "_scale" (entlast) "" '(0 0 0) radius)
  (cmd-SW))

(defun C:DODECAHEDRON  (/ radius vertices faces edges)
  (setq radius   (getdist '(0 0 0) "\nCircumscribed sphere radius: ")
        vertices '((0.57735 -0.187592 0.794654)
                   (0.356822 0.491123 0.794654)
                   (-0.356822 0.491123 0.794654)
                   (-0.57735 -0.187592 0.794654)
                   (-1.11022e-016 -0.607062 0.794654)
                   (0.934172 -0.303531 0.187592)
                   (0.934172 0.303531 -0.187592)
                   (0.57735 0.794654 0.187592)
                   (-5.55112e-017 0.982247 -0.187592)
                   (-0.57735 0.794654 0.187592)
                   (-0.934172 0.303531 -0.187592)
                   (-0.934172 -0.303531 0.187592)
                   (-0.57735 -0.794654 -0.187592)
                   (5.55112e-017 -0.982247 0.187592)
                   (0.57735 -0.794654 -0.187592)
                   (0.356822 -0.491123 -0.794654)
                   (0.57735 0.187592 -0.794654)
                   (-1.33234e-016 0.607062 -0.794654)
                   (-0.57735 0.187592 -0.794654)
                   (-0.356822 -0.491123 -0.794654))
        faces    '((0 1 2 3 4)
                   (0 5 6 7 1)
                   (1 7 8 9 2)
                   (2 9 10 11 3)
                   (3 11 12 13 4)
                   (4 13 14 5 0)
                   (5 14 15 16 6)
                   (7 6 16 17 8)
                   (9 8 17 18 10)
                   (11 10 18 19 12)
                   (13 12 19 15 14)
                   (15 19 18 17 16))
        edges    '(0  1  1  2  2  3  3  4  4  0  0  5  5  6  6  7
                   7  1  7  8  8  9  9  2  9  10 10 11 11 3  11 12
                   12 13 13 4  13 14 14 5  14 15 15 16 16 6  16 17
                   17 8  17 18 18 10 18 19 19 12 19 15))
  (ent-mesh vertices faces edges)
  (command "_scale" (entlast) "" '(0 0 0) radius)
  (cmd-SW))

;;; Extraction of multiple values from an association list.
(defun values  (key lst / sublist result)
  (while (setq sublist (assoc key lst))
    (setq result (cons (cdr sublist) result)
          lst    (cdr (member sublist lst))))
  (reverse result))

;;; Modifying vertex Z coordinates through entmod
(defun entmod-Z
       (ename equation / ent-list vertices x y z vertices-mod)
  (setq ent-list (entget ename)
        vertices (values 10 ent-list))
  (foreach vertex  vertices
    (setq x            (nth 0 vertex)
          y            (nth 1 vertex)
          z            (apply equation (list x y))
          vertices-mod (cons (list x y z) vertices-mod)))
  (setq vertices-mod (reverse vertices-mod))
  (setq i 0)
  (while (< i (length vertices))
    (setq ent-list (subst (cons 10 (nth i vertices-mod))
                          (cons 10 (nth i vertices))
                          ent-list)
          i        (1+ i)))
  (entmod ent-list))

;;; Modifying vertex Z coordinates through ActiveX (only WINDOWS)
(defun ax-mod-z  (ename equation / obj i pt)
  (vl-load-com)
  (setq obj (vlax-ename->vla-object ename)
        i   0)
  (repeat (vla-get-VertexCount obj)
    (setq
      pt (vlax-safearray->list
           (vlax-variant-value (vla-get-Coordinate obj i))))
    (vla-put-Coordinate
      obj
      i
      (vlax-3d-point
        (list (nth 0 pt)
              (nth 1 pt)
              (apply equation (list (nth 0 pt) (nth 1 pt))))))
    (setq i (1+ i))))

;;;--------------------------------------------------------------

;;; Universal mesh vertex modifying function
(defun entmod-XYZ  (ename equation offset axis / ent-list vertices i j k lst-mod
                    height scale-f v-i v-j v-mod vertices-mod)
  (setq ent-list (entget ename)
        vertices (values 10 ent-list))
  (cond ((= axis 0) (setq i 1 j 2))
        ((= axis 1) (setq i 0 j 2))
        ((= axis 2) (setq i 0 j 1)))
  (setq lst-mod
         (mapcar
           '(lambda (vert) (apply equation (list (nth i vert) (nth j vert))))
           vertices)
        height  (- (apply 'max lst-mod) (apply 'min lst-mod))
        scale-f (/ offset height))
  (foreach vertex  vertices
    (setq v-i   (nth i vertex)
          v-j   (nth j vertex)
          v-mod (+ (nth axis vertex) (* scale-f (apply equation (list v-i v-j)))))
    (cond ((= axis 0) (setq vertices-mod (cons (list v-mod v-i v-j) vertices-mod)))
          ((= axis 1) (setq vertices-mod (cons (list v-i v-mod v-j) vertices-mod)))
          ((= axis 2) (setq vertices-mod (cons (list v-i v-j v-mod) vertices-mod)))))
  (setq vertices-mod (reverse vertices-mod))
  (setq k 0)
  (while (< k (length vertices))
    (setq ent-list
           (subst (cons 10 (nth k vertices-mod)) (cons 10 (nth k vertices)) ent-list)
          k        (1+ k)))
  (entmod ent-list)
  (entupd ename))

;;; Mesh shaping program
(defun C:SHAPE-M  (/ ent axis equation offset)
  (command "_undo" "_begin")
  (prompt "\nSelect MESH to shape: ")
  (while (not (setq ent (ssget "_:S" '((0 . "MESH")))))
    (prompt "\nSelect MESH to shape: "))
  (initget (+ 2 4) "X Y Z")
  (setq axis
         (getkword "\nSpecify shaping axis [X/Y/Z]<Z>"))
  (cond ((= axis "X") (setq axis 0))
        ((= axis "Y") (setq axis 1))
        (t (setq axis 2)))
  (setq ent (ssname ent 0))
  (initget 1)
  (setq equation
         (getstring
           "\nSpecify shaping equation function: "))
  (cond
    ((car (atoms-family 0 (list equation)))
     (initget (+ 1 2 4))
     (setq offset (getdist "\nShaping offset: "))
     (entmod-xyz ent (read equation) offset axis))
    (t
     (prompt
       (strcat
         "\nFunction "
         equation
         " has not been defined in the current context."))))
  (command "_undo" "_end")
  (princ))


;;; SURFACE EQUATIONS:

(defun s01 (x y) (* x y))

(defun s02 (x y /)
  (sqrt (abs (* x y))))

(defun s03 (x y)
  (+ (expt x 2) (expt y 2)))

(defun s04 (x y) ;Hyperbolic paraboloid
  (- (expt x 2) (expt y 2)))

(defun s05  (x y /) ;Revolution paraboloid
  (- (+ (expt x 2)
        (expt y 2))))

(defun s06  (x y)
  (- (+ (expt x 2) (expt y 2) (expt y 3)) (expt x 3)))

(defun s07 (x y /)
  (exp (- (+ (expt x 2) (expt y 2)))))

(defun s08  (x y /)
  (/ (sin (* pi x y)) (+ 0.5 (expt x 2) (exp 2))))

(defun s09  (x y /)
  (cos (* x y)))

(defun s10 (x y /)
  (sin (+ (expt x 2) (expt y 2))))


(prompt "\nAU2012-MESH.LSP loaded")(princ)
